"""
physical_models_vec.py

    A module for material strength behavior to be imported into python scripts for
    optimizaton or training emulators.  Adapted from strength_models_add_ptw.py

    Authors:
        DJ Luscher,    djl@lanl.gov
        Peter Trubey,  ptrubey@lanl.gov
        Devin Francom, dfrancom@lanl.gov
        JeeYeon Plohr, jplohr@lanl.gov
        Sky Sjue, sjue@lanl.gov
        Lauren VanDervort, @lvandervort@lanl.gov
        Daniel N Blaschke, dblaschke@lanl.gov
"""

import math
from math import pi

import numpy as np

np.seterr(all="raise")

try:
    from numba import jit

    @jit(nopython=True)
    def erf(x):
        """numba.jit-compatible erf."""
        out = np.empty(x.shape)
        for i, xi in enumerate(x):
            out[i] = math.erf(xi)
        return out

except ImportError:
    from functools import partial

    from scipy.special import erf

    def jit(func=None, forceobj=True, nopython=False):
        """Dummy decorator if numba is unavailable at runtime."""
        return func or partial(jit, forceobj=forceobj, nopython=nopython)

## Error Definitions


class ConstraintError(ValueError):
    pass


class PTWStressError(FloatingPointError):
    pass


## Model Definitions


class BaseModel:
    """
    Base Class for property Models (flow stress, specific heat, melt, density,
    etc.).  Must be instantiated as a child of MaterialModel
    """

    params = []
    consts = []

    def value(self, *args):
        pass

    def update_parameters(self, x):
        self.parent.parameters.update_parameters(x, self.params)

    def __init__(self, parent):
        self.parent = parent


########################
# Specific Heat Models
########################


class Constant_Specific_Heat(BaseModel):
    """
    Constant Specific Heat Model
    """

    consts = ["Cv0"]

    def value(self, *args):
        return self.parent.parameters.Cv0 * np.ones(len(self.parent.state.T))


# class Linear_Specific_Heat(BaseModel):
#    """
#    Linear Specific Heat Model
#    """
#    consts = ['Cv0', 'T0', 'dCdT']

#    def value(self, *args):
#        c0=self.parent.parameters.Cv0
#        t0=self.parent.parameters.T0
#        dcdt=self.parent.parameters.dCdT
#        tnow=self.parent.state.T
#        cnow=c0+(tnow-t0)*dcdt
#        return cnow


class Linear_Specific_Heat(BaseModel):
    """
    Linear Specific Heat Model
    """

    consts = ["c0", "c1"]

    def value(self, *args):
        tnow = self.parent.state.T
        cnow = self.parent.parameters.c0 + self.parent.parameters.c1 * tnow
        return cnow


class Quadratic_Specific_Heat(BaseModel):
    """
    Quadratic Specific Heat Model
    """

    consts = ["c0", "c1", "c2"]

    def value(self, *args):
        tnow = self.parent.state.T
        cnow = (
            self.parent.parameters.c0
            + self.parent.parameters.c1 * tnow
            + self.parent.parameters.c2 * tnow**2
        )
        return cnow


class Cubic_Specific_Heat(BaseModel):
    """
    Cubic Specific Heat Model
    """

    consts = ["c0", "c1", "c2", "c3"]

    def value(self, *args):
        tnow = self.parent.state.T
        cnow = (
            self.parent.parameters.c0
            + self.parent.parameters.c1 * tnow
            + self.parent.parameters.c2 * tnow**2
            + self.parent.parameters.c3 * tnow**3
        )
        return cnow


class Piecewise_Linear_Specific_Heat(BaseModel):
    """
    Piecewise Linear Specific Heat Model
    Cv (T) = c0_0 + c1_0 * T for T<=T_t
    Cv (T) = c0_1 + c1_1 * T for T>T_t
    """

    consts = ["T_t", "c0_0", "c1_0", "c0_1", "c1_1"]

    def value(self, *args):
        tnow = self.parent.state.T
        intercept = np.repeat(self.parent.parameters.c0_0, len(tnow))
        slope = np.repeat(self.parent.parameters.c1_0, len(tnow))
        intercept[np.where(tnow > self.parent.parameters.T_t)] = (
            self.parent.parameters.c0_1
        )
        slope[np.where(tnow > self.parent.parameters.T_t)] = (
            self.parent.parameters.c1_1
        )
        cnow = intercept + slope * tnow
        return cnow


class Piecewise_Quadratic_Specific_Heat(BaseModel):
    """
    Piecewise Quadratic Specific Heat Model
    Cv (T) = c0_0 + c1_0 * T + c2_0 * T**2 for T<=T_t
    Cv (T) = c0_1 + c1_1 * T + c2_1 * T**2 for T>T_t
    """

    consts = ["T_t", "c0_0", "c1_0", "c2_0", "c0_1", "c1_1", "c2_1"]

    def value(self, *args):
        tnow = self.parent.state.T
        pow_0_coeff = np.repeat(self.parent.parameters.c0_0, len(tnow))
        pow_1_coeff = np.repeat(self.parent.parameters.c1_0, len(tnow))
        pow_2_coeff = np.repeat(self.parent.parameters.c2_0, len(tnow))

        pow_0_coeff[np.where(tnow > self.parent.parameters.T_t)] = (
            self.parent.parameters.c0_1
        )
        pow_1_coeff[np.where(tnow > self.parent.parameters.T_t)] = (
            self.parent.parameters.c1_1
        )
        pow_2_coeff[np.where(tnow > self.parent.parameters.T_t)] = (
            self.parent.parameters.c2_1
        )

        cnow = pow_0_coeff + pow_1_coeff * tnow + pow_2_coeff * tnow * tnow
        return cnow


class Piecewise_Cubic_Specific_Heat(BaseModel):
    """
    Cubic Specific Heat Model
    Piecewise Quadratic Specific Heat Model
    Cv (T) = c0_0 + c1_0 * T + c2_0 * T**2  + c3_0 * T**3 for T<=T_t
    Cv (T) = c0_1 + c1_1 * T + c2_1 * T**2  + c3_1 * T**3 for T>T_t
    """

    consts = [
        "T_t",
        "c0_0",
        "c1_0",
        "c2_0",
        "c3_0",
        "c0_1",
        "c1_1",
        "c2_1",
        "c3_1",
    ]

    def value(self, *args):
        tnow = self.parent.state.T
        pow_0_coeff = np.repeat(self.parent.parameters.c0_0, len(tnow))
        pow_1_coeff = np.repeat(self.parent.parameters.c1_0, len(tnow))
        pow_2_coeff = np.repeat(self.parent.parameters.c2_0, len(tnow))
        pow_3_coeff = np.repeat(self.parent.parameters.c3_0, len(tnow))

        pow_0_coeff[np.where(tnow > self.parent.parameters.T_t)] = (
            self.parent.parameters.c0_1
        )
        pow_1_coeff[np.where(tnow > self.parent.parameters.T_t)] = (
            self.parent.parameters.c1_1
        )
        pow_2_coeff[np.where(tnow > self.parent.parameters.T_t)] = (
            self.parent.parameters.c2_1
        )
        pow_3_coeff[np.where(tnow > self.parent.parameters.T_t)] = (
            self.parent.parameters.c3_1
        )

        cnow = (
            pow_0_coeff
            + pow_1_coeff * tnow
            + pow_2_coeff * tnow * tnow
            + pow_3_coeff * tnow * tnow * tnow
        )
        return cnow


########################
# Density Models
########################


class Constant_Density(BaseModel):
    """
    Constant Density Model
    """

    consts = ["rho0"]

    def value(self, *args):
        return self.parent.parameters.rho0 * np.ones(len(self.parent.state.T))


# class Linear_Density(BaseModel):
#    """
#    Linear Density Model
#    """
#    consts = ['rho0', 'T0', 'dRhodT']

#    def value(self, *args):
#        r0=self.parent.parameters.rho0
#        t0=self.parent.parameters.T0
#        drdt=self.parent.parameters.dRhodT
#        tnow=self.parent.state.T
#        rnow=r0+drdt*(tnow-t0)
#        return rnow


class Linear_Density(BaseModel):
    """
    Linear Density Model
    """

    consts = ["r0", "r1"]

    def value(self, *args):
        tnow = self.parent.state.T
        rnow = self.parent.parameters.r0 + self.parent.parameters.r1 * tnow
        return rnow


class Quadratic_Density(BaseModel):
    """
    Quadratic Density Model
    """

    consts = ["r0", "r1", "r2"]

    def value(self, *args):
        tnow = self.parent.state.T
        rnow = (
            self.parent.parameters.r0
            + self.parent.parameters.r1 * tnow
            + self.parent.parameters.r2 * tnow**2
        )
        return rnow


class Cubic_Density(BaseModel):
    """
    Quadratic Density Model
    """

    consts = ["r0", "r1", "r2", "r3"]

    def value(self, *args):
        tnow = self.parent.state.T
        rnow = (
            self.parent.parameters.r0
            + self.parent.parameters.r1 * tnow
            + self.parent.parameters.r2 * tnow**2
            + self.parent.parameters.r3 * tnow**3
        )
        return rnow


########################
# Melt Temperature Models
########################


class Constant_Melt_Temperature(BaseModel):
    """
    Constant Melt Temperature Model
    """

    consts = ["Tmelt0"]

    def value(self, *args):
        return self.parent.parameters.Tmelt0 * np.ones(len(self.parent.state.T))


# class Linear_Melt_Temperature(BaseModel):
#    """
#    Linear Melt Temperature Model
#    """
#    consts=['Tmelt0', 'rho0', 'dTmdRho']#

#    def value(self, *args):
#        tm0=self.parent.parameters.Tmelt0
#        rnow=self.parent.state.rho
#        dtdr=self.parent.parameters.dTmdRho
#        r0=self.parent.parameters.rho0
#        tmeltnow=tm0+dtdr*(rnow-r0)
#        return tmeltnow


class Linear_Melt_Temperature(BaseModel):
    """
    Linear Melt Temperature Model
    """

    consts = ["tm0", "tm1"]

    def value(self, *args):
        rnow = self.parent.state.rho

        tmeltnow = (
            self.parent.parameters.tm0 + self.parent.parameters.tm1 * rnow
        )
        return tmeltnow


class Quadratic_Melt_Temperature(BaseModel):
    """
    Quadratic Melt Temperature Model
    """

    consts = ["tm0", "tm1", "tm2"]

    def value(self, *args):
        rnow = self.parent.state.rho
        tmeltnow = (
            self.parent.parameters.tm0
            + self.parent.parameters.tm1 * rnow
            + self.parent.parameters.tm2 * rnow**2
        )
        return tmeltnow


class Cubic_Melt_Temperature(BaseModel):
    """
    Cubic Melt Temperature Model
    """

    consts = ["tm0", "tm1", "tm2", "tm3"]

    def value(self, *args):
        rnow = self.parent.state.rho
        tmeltnow = (
            self.parent.parameters.tm0
            + self.parent.parameters.tm1 * rnow
            + self.parent.parameters.tm2 * rnow**2
            + self.parent.parameters.tm3 * rnow**3
        )
        # print(tmeltnow)
        return tmeltnow


class BGP_Melt_Temperature(BaseModel):
    consts = ["Tm_0", "rho_m", "gamma_1", "gamma_3", "q3"]

    def value(self, *args):
        mp = self.parent.parameters
        rho = self.parent.state.rho

        melt_temp = (
            mp.Tm_0
            * np.power(rho / mp.rho_m, 1.0 / 3.0)
            * np.exp(
                6
                * mp.gamma_1
                * (np.power(mp.rho_m, -1.0 / 3.0) - np.power(rho, -1.0 / 3.0))
                + 2.0
                * mp.gamma_3
                / mp.q3
                * (np.power(mp.rho_m, -mp.q3) - np.power(rho, -mp.q3))
            )
        )
        return melt_temp


# Shear Modulus Models


class Constant_Shear_Modulus(BaseModel):
    consts = ["G0"]

    def value(self, *args):
        return self.parent.parameters.G0 * np.ones(len(self.parent.state.T))


# class Linear_Shear_Modulus(BaseModel):
#    consts =  ['G0', 'rho0', 'dGdRho' ]#

#    def value(self, *args):
#         g0=self.parent.parameters.G0
#         rho0=self.parent.parameters.rho0
#         dgdr=self.parent.parameters.dGdRho
#         rnow=self.parent.state.rho
#         gnow=g0+dgdr*(rnow-rho0)
#         return gnow


class Linear_Cold_PW_Shear_Modulus(BaseModel):
    consts = ["g0", "g1", "alpha"]

    def value(self, *args):
        mp = self.parent.parameters
        rho = self.parent.state.rho
        temp = self.parent.state.T
        tmelt = self.parent.state.Tmelt
        cold_shear = mp.g0 + mp.g1 * rho
        gnow = cold_shear * (1.0 - mp.alpha * (temp / tmelt))

        gnow[np.where(temp >= tmelt)] = 0.0
        gnow[np.where(gnow < 0)] = 0.0

        return gnow


class Quadratic_Cold_PW_Shear_Modulus(BaseModel):
    consts = ["g0", "g1", "g2", "alpha"]

    def value(self, *args):
        mp = self.parent.parameters
        rho = self.parent.state.rho
        temp = self.parent.state.T
        tmelt = self.parent.state.Tmelt
        cold_shear = mp.g0 + mp.g1 * rho + mp.g2 * rho**2
        gnow = cold_shear * (1.0 - mp.alpha * (temp / tmelt))

        gnow[np.where(temp >= tmelt)] = 0.0
        gnow[np.where(gnow < 0)] = 0.0

        return gnow


class Simple_Shear_Modulus(BaseModel):
    consts = ["G0", "alpha"]

    def value(self, *args):
        mp = self.parent.parameters
        temp = self.parent.state.T
        tmelt = self.parent.state.Tmelt

        return mp.G0 * (1.0 - mp.alpha * (temp / tmelt))


@jit(nopython=True)
def _BGP_PW_Shear_Modulus(
    rho, G0, rho_0, gamma_1, gamma_2, q2, alpha, temp, tmelt
):
    """BPG model provides cold shear, i.e. shear modulus at zero temperature as a function of density.
    PW describes the (linear) temperature dependence of the shear modulus. (Same dependency as
    in Simple_Shear_modulus.)
    With these two models combined, we get the shear modulus as a function of density and temperature;
    see Burakovsky, Greeff, Preston, Phys. Rev. B67 (2003) 094107, DOI:10.1103/PhysRevB.67.094107"""
    cold_shear = (
        G0
        * np.power(rho / rho_0, 4.0 / 3.0)
        * np.exp(
            6.0 * gamma_1 * (1 / np.cbrt(rho_0) - 1 / np.cbrt(rho))
            + 2 * gamma_2 / q2 * (np.power(rho_0, -q2) - np.power(rho, -q2))
        )
    )
    gnow = cold_shear * (1.0 - alpha * (temp / tmelt))

    gnow[np.where(temp >= tmelt)] = 0.0
    gnow[np.where(gnow < 0)] = 0.0

    # if temp >= tmelt: gnow = 0.0
    # if gnow < 0.0:    gnow = 0.0
    return gnow


class BGP_PW_Shear_Modulus(BaseModel):
    """BPG model provides cold shear, i.e. shear modulus at zero temperature as a function of density.
    PW describes the (linear) temperature dependence of the shear modulus. (Same dependency as
    in Simple_Shear_modulus.)
    With these two models combined, we get the shear modulus as a function of density and temperature;
    see Burakovsky, Greeff, Preston, Phys. Rev. B67 (2003) 094107, DOI:10.1103/PhysRevB.67.094107"""

    consts = ["G0", "rho_0", "gamma_1", "gamma_2", "q2", "alpha"]

    def value(self, *args):
        mp = self.parent.parameters
        rho = self.parent.state.rho
        temp = self.parent.state.T
        tmelt = self.parent.state.Tmelt
        gnow = _BGP_PW_Shear_Modulus(
            rho=rho,
            G0=mp.G0,
            rho_0=mp.rho_0,
            gamma_1=mp.gamma_1,
            gamma_2=mp.gamma_2,
            q2=mp.q2,
            alpha=mp.alpha,
            temp=temp,
            tmelt=tmelt,
        )
        return gnow


class Stein_Shear_Modulus(BaseModel):
    # consts = ['G0', 'sgA', 'sgB']
    # assuming constant density and pressure
    # so we only include the temperature dependence
    consts = ["G0", "sgB"]
    eta = 1.0

    def value(self, *args):
        mp = self.parent.parameters
        temp = self.parent.state.T
        tmelt = self.parent.state.Tmelt
        # just putting this here for completeness
        # aterm = a/eta**(1.0/3.0)*pressure
        aterm = 0.0
        bterm = mp.sgB * (temp - 300.0)
        gnow = mp.G0 * (1.0 + aterm - bterm)
        # if temp >= tmelt: gnow = 0.0
        # if gnow < 0.0:    gnow = 0.0
        gnow[np.where(temp >= tmelt)] = 0.0
        gnow[np.where(gnow < 0)] = 0.0
        return gnow


# Yield Stress Models


class Constant_Yield_Stress(BaseModel):
    """
    Constant Yield Stress Model
    """

    consts = ["yield_stress", "chi"]

    def value(self, *args):
        return self.parent.parameters.yield_stress * np.ones(
            len(self.parent.state.T)
        )


# def fast_pow(a, b):
#     """
#     Numpy power is slow, this is faster.  Gets a**b for a and b np arrays.
#     """
#     ## no longer true in numpy 1.24 (=our requirement in toml file) and higher; deprecate / remove?
#     cond = a > 0
#     out = a * 0.0
#     out[cond] = np.exp(b[cond] * np.log(a[cond]))
#     return out
fast_pow = np.power ## do we need this?


def pos(a):
    return np.maximum(0, a)


class JC_Yield_Stress(BaseModel):
    params = ["A", "B", "C", "n", "m"]
    consts = ["Tref", "edot0", "chi"]

    def value(self, edot):
        mp = self.parent.parameters
        eps = self.parent.state.strain
        t = self.parent.state.T
        tmelt = self.parent.state.Tmelt

        # th = np.max([(t - mp.Tref) / (tmelt - mp.Tref), 0.])
        th = pos((t - mp.Tref) / (tmelt - mp.Tref))

        Y = (
            (mp.A + mp.B * np.power(eps, mp.n))
            * (1.0 + mp.C * np.log(edot / mp.edot0))
            * (1.0 - np.power(th, mp.m))
        )
        return Y


@jit(nopython=True)
def _PTW_corefct(
    p,
    kappa,
    s0,
    sInf,
    y0,
    yInf,
    y1,
    y2,
    beta,
    theta,
    lgamma,
    xiDot,
    edot,
    t_hom,
    eps,
    small=1.0e-10,
):
    """jit-compiled subroutine of the PTW_Yield_Stress class"""
    log_xid_ed = np.log(xiDot / edot)
    argErf = kappa * t_hom * (lgamma + log_xid_ed)
    Erfres = erf(argErf)

    saturation1 = s0 - (s0 - sInf) * Erfres
    saturation2 = s0 * np.exp(beta * (-lgamma - log_xid_ed))
    sat_cond = saturation1 > saturation2
    tau_s = np.copy(saturation2)
    tau_s[np.where(sat_cond)] = saturation1[sat_cond]

    ayield = y0 - (y0 - yInf) * Erfres
    byield = y1 * np.exp(-y2 * (lgamma + log_xid_ed))
    cyield = s0 * np.exp(-beta * (lgamma + log_xid_ed))

    y_cond = byield < cyield
    dyield = np.copy(cyield)
    dyield[np.where(y_cond)] = byield[y_cond]

    y_cond2 = ayield > dyield
    tau_y = np.copy(dyield)
    tau_y[np.where(y_cond2)] = ayield[y_cond2]
    scaled_stress = tau_s
    ind = np.where((p > small) * (np.abs(tau_s - tau_y) > small))
    eArg1 = (p * (tau_s - tau_y) / (s0 - tau_y))[ind]
    eArg2 = (
        (eps * p * theta)[ind] / (s0 - tau_y)[ind] / (np.exp(eArg1) - 1.0)
    )  # eArg1 already subsetted by ind
    if np.any((1.0 - (1.0 - np.exp(-eArg1)) * np.exp(-eArg2)) <= 0) or np.any(
        np.isinf(1.0 - (1.0 - np.exp(-eArg1)) * np.exp(-eArg2))
    ):
        print("bad")
    theLog = np.log(1.0 - (1.0 - np.exp(-eArg1)) * np.exp(-eArg2))
    scaled_stress[ind] = tau_s[ind] + (s0[ind] - tau_y[ind]) * theLog / p[ind]
    ind2 = np.where((p <= small) * (tau_s > tau_y))
    scaled_stress[ind2] = +tau_s[ind2] - (tau_s - tau_y)[ind2] * np.exp(
        -eps[ind2] * theta[ind2] / (tau_s - tau_y)[ind2]
    )
    return scaled_stress


@jit(nopython=True)
def PTW_goodparam(s0, sInf, y0, yInf, y1, y2, beta):
    """checks if the given PTW parameter set is valid"""
    return (
        (sInf < s0)
        * (yInf < y0)
        * (y0 < s0)
        * (yInf < sInf)
        * (y1 > s0)
        * (y2 > beta)
    )


class PTW_Yield_Stress(BaseModel):
    """This class implements the PTW flow stress model"""

    params = [
        "theta",
        "p",
        "s0",
        'beta',
        "sInf",
        "kappa",
        "lgamma",
        "y0",
        "yInf",
        "y1",
        "y2",
    ]
    consts = ["rho0", "matomic", "chi"]

    # @profile
    def value(self, edot):
        """
        function used to define PTW flow stress model
        arguments are:
            - edot: scalar, strain rate
            - material: an instance of MaterialModel class
        returns the flow stress at the current material state
        and specified strain rate
        """
        mp = self.parent.parameters
        eps = self.parent.state.strain
        temp = self.parent.state.T
        tmelt = self.parent.state.Tmelt
        shear = self.parent.state.G

        good = PTW_goodparam(
            mp.s0, mp.sInf, mp.y0, mp.yInf, mp.y1, mp.y2, mp.beta
        )
        if np.any(np.logical_not(good)):
            # return np.array([-999.]*len(good))
            raise ConstraintError("PTW bad val")

        # convert to 1/s strain rate since PTW rate is in that unit
        edot = edot * 1.0e6

        t_hom = temp / tmelt
        # this one is commented because it is assumed that
        # the material state computes the temperature dependence of
        # the shear modulus
        # shear = shear * (1.0 - mp.alpha * t_hom)
        # print("ptw shear is "+str(shear))

        afact = (4.0 / 3.0) * pi * mp.rho0 / mp.matomic
        # ainv is 1/a where 4/3 pi a^3 is the atomic volume
        ainv = afact ** (1.0 / 3.0)

        # transverse wave velocity up to units
        xfact = np.sqrt(shear / mp.rho0)
        # PTW characteristic strain rate [ 1/s ]
        xiDot = 0.5 * ainv * xfact * pow(6.022e29, 1.0 / 3.0) * 1.0e4

        # should be flow stress in units of Mbar
        out = (
            _PTW_corefct(
                mp.p,
                mp.kappa,
                mp.s0,
                mp.sInf,
                mp.y0,
                mp.yInf,
                mp.y1,
                mp.y2,
                mp.beta,
                mp.theta,
                mp.lgamma,
                xiDot,
                edot,
                t_hom,
                eps,
                small=1.0e-10,
            )
            * shear
            * 2.0
        )
        out[np.where(np.logical_not(good))] = -999.0
        return out


class Stein_Flow_Stress(BaseModel):
    params = ["y0", "a", "b", "beta", "n", "ymax"]
    consts = ["G0", "epsi", "chi"]

    def value(self, *args):
        mp = self.parent.parameters
        temp = self.parent.state.T
        tmelt = self.parent.state.Tmelt
        shear = self.parent.state.G
        eps = self.parent.state.strain
        fnow = np.power((1.0 + mp.beta * (mp.epsi + eps)), mp.n)

        cond1 = fnow * mp.y0 > mp.ymax
        fnow[cond1] = (mp.ymax / mp.y0)[cond1]
        cond2 = temp > tmelt
        fnow[cond2] = 0.0

        # if fnow*mp.y0 > mp.ymax: fnow = mp.ymax/mp.y0
        # if temp > tmelt: fnow = 0.0
        return mp.y0 * fnow * shear / mp.G0


## Parameters Definition


class ModelParameters:
    params = []
    consts = []
    parent = None

    def update_parameters(self, x):
        if isinstance(x, np.ndarray):
            self.__dict__.update(dict(zip(self.params, x)))
        elif isinstance(x, dict):
            for key in self.params:
                self.__dict__[key] = x[key]
        elif isinstance(x, list):
            assert len(x) == len(self.params), "Incorrect number of parameters!"
            for i, xi in enumerate(self.params):
                self.__dict__[self.params[i]] = xi
        else:
            raise ValueError(f"Type {type(x)} is not supported.")

    def __init__(self, parent):
        self.parent = parent


## State Definition


class MaterialState:
    T = None
    Tmelt = None
    stress = None
    strain = None
    G = None

    def set_state(self, T=300.0, strain=0.0, stress=0.0):
        self.T = T
        self.strain = strain
        self.stress = stress

    def __init__(self, parent, T=300.0, strain=0.0, stress=0.0):
        self.parent = parent
        self.set_state(T, strain, stress)


## Material Model Definition


class MaterialModel:
    def __init__(
        self,
        parameters=ModelParameters,
        initial_state=MaterialState,
        flow_stress_model=Constant_Yield_Stress,
        specific_heat_model=Constant_Specific_Heat,
        shear_modulus_model=Constant_Shear_Modulus,
        melt_model=Constant_Melt_Temperature,
        density_model=Constant_Density,
    ):
        """
        Initialization routine for Material Model.  All of the arguments
        supplied are classes, which are then instantiated within the function.

        The reason for doing this is that then we can pass the MaterialModel
        instance to the physical models so that the model's parent can be
        declared at instantiation.  then the model.value() function can reach
        into the parent class to find whatever it needs.
        """
        self.parameters = parameters(self)
        self.state = initial_state(self)

        self.flow_stress = flow_stress_model(self)
        self.specific_heat = specific_heat_model(self)
        self.shear_modulus = shear_modulus_model(self)
        self.melt_model = melt_model(self)
        self.density = density_model(self)

        params = (
            self.flow_stress.params
            + self.specific_heat.params
            + self.shear_modulus.params
            + self.melt_model.params
            + self.density.params
        )
        consts = set(
            self.flow_stress.consts
            + self.specific_heat.consts
            + self.shear_modulus.consts
            + self.melt_model.consts
            + self.density.consts
        )

        assert len(set(params)) == len(params) , "Some Duplicate Parameters between models"
        assert len(set(params).intersection(set(consts))) == 0 , "Duplicate item in parameters and constants"

        self.parameters.params = params
        self.parameters.consts = consts

    def get_parameter_list(
        self,
    ):
        """
        The list of parameters used in the model.
        This also describes the order of their appearance in the sampling results
        """
        return self.parameters.params

    def get_constants_list(
        self,
    ):
        """
        List of Constants used in the model
        """
        return self.parameters.consts

    def update_state(self, edot, dt):
        chi = self.parameters.chi
        self.state.Cv = self.specific_heat.value()
        self.state.rho = self.density.value()
        # if we are working with microseconds, then this is a reasonable value
        # if we work in seconds, it should be changed to ~1.
        edotcrit = 1.0e-6
        # if edot > edotcrit:
        #  self.state.T += chi * self.state.stress * edot * dt / (self.state.Cv * self.state.rho)
        cond = edot > edotcrit
        # if any(cond):
        self.state.T = (
            self.state.T
            + cond
            * chi
            * self.state.stress
            * edot
            * dt
            / (self.state.Cv * self.state.rho)
        )
        self.state.strain = self.state.strain + edot * dt

        self.state.Tmelt = self.melt_model.value()
        self.state.G = self.shear_modulus.value()
        self.state.stress = self.flow_stress.value(edot)

    def update_parameters(self, x):
        self.parameters.update_parameters(x)

    def initialize(self, parameters, constants):
        """
        Initialize the model at a given set of parameters, constants
        """
        ## if user assumed one or more parameters constant, they would be in the constants var instead;
        ## check for this first:
        if not isinstance(self.parameters.params, list):
            self.parameters.params = list(self.parameters.params)
        if not isinstance(self.parameters.consts, set):
            self.parameters.consts = set(self.parameters.consts)
        user_constants = set(self.parameters.params).difference(parameters)
        for usercnst in user_constants:
            self.parameters.consts |= {usercnst}
        try:
            self.parameters.__dict__.update(
                {
                    key: parameters[key]
                    for key in set(self.parameters.params).difference(
                        user_constants
                    )
                },
            )
            self.parameters.__dict__ |= {
                key: constants[key] for key in user_constants
            }
        except KeyError:
            print(
                "{} missing from list of supplied parameters".format(
                    set(self.parameters.params).difference(
                        set(parameters.keys())
                    )
                )
            )
            raise
        try:
            self.parameters.__dict__.update(
                {key: constants[key] for key in self.parameters.consts},
            )
        except KeyError:
            print(
                "{} missing from list of supplied constants".format(
                    set(self.parameters.consts).difference(
                        set(constants.keys())
                    )
                )
            )
            raise

    def initialize_state(self, T=300.0, stress=0.0, strain=0.0):
        self.state.set_state(T, stress, strain)

    def set_history_variables(self, emax, edot, Nhist):
        self.emax = emax
        self.edot = edot
        self.Nhist = Nhist

    def get_history_variables(self):
        return [self.emax, self.edot, self.Nhist]

    def compute_state_history(self, strain_history):
        strains = strain_history["strains"]
        times = strain_history["times"]
        strain_rate = strain_history["strain_rate"]
        # Nhist = len(strains)
        # nrep = len(self.parameters.kappa)
        nrep, Nhist = strains.shape  # nexp * nhist array

        results = np.empty((Nhist, 6, nrep))

        state = self.state
        self.update_state(strain_rate[:, 0], 0.0)

        # import pdb
        # pdb.set_trace()

        results[0] = np.array([
            times[:, 0],
            state.strain,
            state.stress,
            state.T,
            state.G,
            state.rho,
        ])  # np.repeat(state.rho,nrep)])

        for i in range(1, Nhist):
            self.update_state(
                strain_rate[:, i - 1], times[:, i] - times[:, i - 1]
            )
            # self.update_state(strain_rate.T[i-1], times.T[i] - times.T[i-1])
            # results[i] = [times[i], state.strain, state.stress, state.T, state.G, state.rho]
            results[i] = np.array([
                times[:, i],
                state.strain,
                state.stress,
                state.T,
                state.G,
                state.rho,
            ])  # np.repeat(state.rho, nrep)])

        return results


## function to generate strain history to calculate along
##  Should probably make this a method of MaterialModel class


def generate_strain_history(emax, edot, Nhist):
    tmax = emax / edot
    strains = np.linspace(0.0, emax, Nhist)
    nrep = len(edot)
    times = np.empty((nrep, Nhist))
    # for i in range(nrep):
    #    times[i,:] = np.linspace(0., tmax[i], Nhist)
    times = np.linspace(0.0, tmax, Nhist)
    strain_rate = np.empty((nrep, Nhist - 1))
    # for i in range(nrep):
    #    strain_rate[i, :] = np.diff(strains) / np.diff(times[i, :])

    strain_diffs = np.diff(strains)

    strain_rate = strain_diffs[:, np.newaxis] / np.diff(times, axis=0)
    return dict((
        ["times", times.T],
        ["strains", strains],
        ["strain_rate", strain_rate.T],
    ))


def generate_strain_history_new(emax, edot, nhist):
    tmax = emax / edot
    strains = np.linspace(0, emax, nhist)  # nhist * nexp
    times = np.linspace(0, tmax, nhist)  # nhist * nexp
    rates = np.diff(strains, axis=0) / np.diff(
        times, axis=0
    )  # (nhist - 1) * nexp
    return {"times": times.T, "strains": strains.T, "strain_rate": rates.T}



import numpy as np
from numpy import ndarray
from scipy.stats import kstest
from scipy.stats import uniform as Uniform

from impala import superCal as sc


class Friedman:
    def __init__(self, grid: ndarray) -> None:
        """
        grid: Grid at which Friedman function is evaluated.
        """
        self.grid = grid

    def __call__(self, theta: ndarray) -> ndarray:
        """
        theta: Friedman function parameters.
        """
        return (
            10 * np.sin(np.pi * self.grid * theta[0])
            + 20 * (theta[1] - 0.5) ** 2
            + 10 * theta[2]
            + 5 * theta[3]
        )


def generate_data(n_features: int, gridsize: int):
    assert n_features >= 4, (
        "Friedman function has 4 parameters, so n_feature must be at least 4."
    )

    grid = np.linspace(0, 1, gridsize)
    friedman = Friedman(grid)

    # True calibration parameter value. Note that the Friedman function uses
    # only the first four elements. So the remaining elements should have
    # uniform posteriors.
    theta = np.random.rand(1, n_features)

    # True observation error standard deviation.
    sigma = 0.1

    mu = np.apply_along_axis(friedman, 1, theta).squeeze()
    yobs = np.random.normal(mu, sigma)
    param_truth = dict(theta=theta, sigma=sigma)

    return yobs, friedman, n_features, param_truth


def test_friedman_fit():
    np.random.seed(0)
    yobs, friedman, n_features, param_truth = generate_data(
        n_features=6, gridsize=50
    )

    # Create bounds for each theta. Names (keys) need to be strings.
    bounds = {str(index): np.array([0, 1]) for index in range(n_features)}

    # Initialize with the only constraints being the bounds.
    setup = sc.CalibSetup(bounds, constraint_func="bounds")

    # Put the Friedman function into the right structure; this could be replaced
    # with an emulator.
    model = sc.ModelF(friedman, input_names=bounds.keys())

    # NOTE: If you have multiple experiments, just call setup.addVecExperiments
    # multiple times.
    setup.addVecExperiments(
        # observation vector
        yobs=yobs,
        # model that predicts a vector
        model=model,
        # yobs error standard deviation estimate (possibly a vector of estimates
        # for different parts of yobs vector).
        sd_est=[1.0],
        # yobs error degrees of freedom (larger means more confidence in
        # sd_est), same shape as sd_est.
        s2_df=[0],
        # if sd_est is a vector of length 3, this is a vector of length
        # len(yobs) with values (0, 1, 2) indicating which sd_est corresponds to
        # which part of yobs.
        s2_ind=[0] * len(yobs),
    )

    # temperature ladder, typically (1 + step) ** np.arange(ntemps)
    setup.setTemperatureLadder(1.05 ** np.arange(20))

    # MCMC number of iterations, and how often to take a decorrelation step.
    setup.setMCMC(nmcmc=15000, decor=100)

    # Pooled calibration (takes less than a minute).
    out = sc.calibPool(setup)

    # out.theta is has shape (num_mcmc_samples, num_temperatures, n_features).
    # Use index zero to get the posterior for the coldest chain (temperatre=0).
    burn = 6000  # burn-in: initial samples to discard.
    thin = 2  # thinning factor: take every 2 samples after burn-in.
    theta_posterior = out.theta[burn::thin, 0]

    # Number of parameters utilized by the Friedman function.
    num_utilized_parameters = 4

    # Test that the posterior for the superfluous parameters in theta are
    # uniform.
    superfluous_theta_posterior = theta_posterior[:, num_utilized_parameters:]
    for i, theta in enumerate(superfluous_theta_posterior.T):
        assert kstest(theta, Uniform().cdf).pvalue > 0.10, (
            f"theta_{i} is not Uniform!"
        )

    # Test that the posterior for the utilized parameters in theta are not
    # uniform.
    utilized_theta_posterior = theta_posterior[:, :num_utilized_parameters]
    for i, theta in enumerate(utilized_theta_posterior.T):
        assert kstest(theta, Uniform().cdf).pvalue < 0.01, (
            f"theta_{i} is Uniform!"
        )

    # Test that the true values of theta are in the 95% credible interval.
    lower = np.quantile(theta_posterior, 0.025, 0)
    upper = np.quantile(theta_posterior, 0.975, 0)
    assert np.all(lower < param_truth["theta"])
    assert np.all(param_truth["theta"] < upper)



import unittest

import numpy as np

from impala.physics import physical_models_vec as pmh


# Constant yield stress model
class TestConstantYieldStress(unittest.TestCase):
    def setUp(self):
        self.params = {}
        self.consts = {
            "yield_stress": 0.1,  # MBar
            "rho0": 8.96,  # g/cc
            "Cv0": 0.383e-5,  # MBar*cc/g*K
            "G0": 0.4578,  # MBar
            "Tmelt0": 1356.0,  # K
            "chi": 1.0,
        }
        self.model_const_y = pmh.MaterialModel()

    def test_isothermal_lowrate(self):
        """
        Const. yield, g0, Tm. Rates less than 1e-6/us (~1/s), should be isothermal T=const
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([1.0]), edot=np.array([1e-7]), nhist=1000
        )
        self.model_const_y.initialize(self.params, self.consts)
        self.model_const_y.initialize_state(T=np.array([298.0]))
        results_const_y = self.model_const_y.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        self.assertEqual(results_const_y[0][3], results_const_y[-1][3])

    def test_adiabatic_highrate(self):
        """
        Const. yield, g0, Tm. Rates greater than 1e-6/us (~1/s), temperature changes adiabatically
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([1.0]), edot=np.array([1e0]), nhist=1000
        )
        self.model_const_y.initialize(self.params, self.consts)
        self.model_const_y.initialize_state(T=np.array([298.0]))
        results_const_y = self.model_const_y.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        self.assertNotEqual(results_const_y[0][3], results_const_y[-1][3])

    def test_constant_stress(self):
        """
        Const. yield, g0, Tm. Yield stress is constant for all strain
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([1.0]), edot=np.array([1e0]), nhist=1000
        )
        self.model_const_y.initialize(self.params, self.consts)
        self.model_const_y.initialize_state(T=np.array([298.0]))
        results_const_y = self.model_const_y.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        for time_i in results_const_y:
            self.assertEqual(self.consts["yield_stress"], time_i[2])



######################################
######################################
### Impala Model Class Definitions ###
######################################
######################################

###############
### Imports ###
###############
import abc
from itertools import cycle

import numpy as np
from scipy.interpolate import interp1d
from scipy.linalg import cho_factor, cholesky

# import pyBASS as pb
# import pyBayesPPR as pbppr
# import physical_models_vec as pm_vec
from impala import physics as pm_vec

########################
### Helper Functions ###
########################


def cor2cov(R, s):  # R is correlation matrix, s is sd vector
    return np.outer(s, s) * R


def chol_sample(mean, cov):
    return mean + np.dot(
        np.linalg.cholesky(cov), np.random.standard_normal(mean.size)
    )


#####################
### Model Classes ### #should have eval method and stochastic attribute
#####################


#######
### AbstractModel: Internal class, not called by users
class AbstractModel:
    """
    Base Class for Simulator/Emulator Models.
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def eval(self, parmat):  # this must be implemented for each model type
        pass

    # @profile
    def llik(self, yobs, pred, cov):  # assumes diagonal cov
        vec = yobs - pred
        vec2 = vec * vec * cov["inv"]
        out = -0.5 * cov["ldet"] - 0.5 * vec2.sum()
        return out

    # @profile
    def lik_cov_inv(self, s2vec):  # default is diagonal covariance matrix
        inv = 1 / s2vec
        ldet = np.log(s2vec).sum()
        out = {"inv": inv, "ldet": ldet}
        return out

    def step(self):
        return


#######
### ModelBassPca_mult: Model with BASS Emulator from pyBASS with Multivariate Output
class ModelBassPca_mult(AbstractModel):
    """
    ModelBassPca_mult: PCA Based BASS Model Emulator for Multivariate Outputs

    ModelBassPca_mult is not recommended for larger-dimensional functional outputs. Instead, use ModelBassPca_func.
    """

    def __init__(self, bmod, input_names, exp_ind=None, s2="MH"):
        """
        bmod        : bassPCA fit
        input_names : list of the names of the inputs to bmod
        s2          : method for handling experiment-specific noise s2; options are 'MH' (Metropolis-Hastings Sampling), 'fix' (fixed at s2_est from addVecExperiments call)
        """
        self.mod = bmod
        self.stochastic = True
        self.nmcmc = len(bmod.bm_list[0].samples.s2)
        self.input_names = input_names
        self.trunc_error_cov = np.cov(self.mod.trunc_error)
        self.basis = self.mod.basis
        self.meas_error_cor = np.eye(self.basis.shape[0])
        self.discrep_cov = np.eye(self.basis.shape[0]) * 1e-12
        self.ii = 0
        npc = self.mod.nbasis
        self.mod_s2 = np.empty([self.nmcmc, npc])
        for i in range(npc):
            self.mod_s2[:, i] = self.mod.bm_list[i].samples.s2
        self.emu_vars = self.mod_s2[self.ii]
        self.yobs = None
        self.marg_lik_cov = None
        self.nd = 0
        if exp_ind is None:
            exp_ind = np.array(0)
        self.nexp = exp_ind.max() + 1
        self.exp_ind = exp_ind
        self.s2 = s2
        self.constants = None
        if s2 == "gibbs":
            raise "Cannot use Gibbs s2 for emulator models."
        return

    def step(self):
        self.ii = np.random.choice(range(self.nmcmc), 1).item()
        self.emu_vars = self.mod_s2[self.ii]
        return

    def eval(self, parmat, pool=None, nugget=False):
        """
        parmat : ~
        """
        parmat_array = np.vstack([
            parmat[v] for v in self.input_names
        ]).T  # get correct subset/ordering of inputs
        pred = self.mod.predict(
            parmat_array, mcmc_use=np.array([self.ii]), nugget=nugget
        )[0, :, :]

        if pool is True:
            return pred
        else:
            nrep = list(parmat.values())[0].shape[0] // self.nexp
            return np.concatenate(
                [
                    pred[
                        np.ix_(
                            np.arange(i, nrep * self.nexp, self.nexp),
                            np.where(self.exp_ind == i)[0],
                        )
                    ]
                    for i in range(self.nexp)
                ],
                1,
            )
            # this is evaluating all experiments for all thetas, which is overkill

    def llik(self, yobs, pred, cov):
        vec = yobs - pred
        out = -0.5 * (cov["ldet"] + vec.T @ cov["inv"] @ vec)
        return out

    def lik_cov_inv(self, s2vec):
        n = len(s2vec)
        Sigma = cor2cov(
            self.meas_error_cor[:n, :n], np.sqrt(s2vec)
        )  # :n is a hack for when ntheta>1 in heir...fix this sometime
        mat = (
            Sigma
            + self.trunc_error_cov
            + self.discrep_cov
            + self.basis @ np.diag(self.emu_vars) @ self.basis.T
        )
        # this doesnt work for vectorized experiments...maybe dont allow those for BASS
        chol = cholesky(mat)
        ldet = 2 * np.sum(np.log(np.diag(chol)))
        # la.dpotri(chol, overwrite_c=True) # overwrites chol with original matrix inverse
        inv = np.linalg.inv(mat)
        out = {"inv": inv, "ldet": ldet}
        return out


#######
### ModelBpprPca_mult: Model with BayesPPR Emulator with Multivariate Output
### Not recommended for larger-dimensional functional outputs: instead, use ModelBpprPca_func
class ModelBpprPca_mult(AbstractModel):
    """
    ModelBpprPca_mult: BayesPPR Model Emulator for Multivariate Outputs

    ModelBpprPca_mult is not recommended for larger-dimensional functional outputs. Instead, use ModelBpprPca_func.
    """

    def __init__(self, bmod, input_names, exp_ind=None, s2="MH"):
        """
        bmod        : bassPCA fit
        input_names : list of the names of the inputs to bmod
        s2          : method for handling experiment-specific noise s2; options are 'MH' (Metropolis-Hastings Sampling), 'fix' (fixed at s2_est from addVecExperiments call)
        """
        self.mod = bmod
        self.stochastic = True
        self.nmcmc = len(bmod.bm_list[0].samples.sdResid)
        self.input_names = input_names
        self.trunc_error_cov = np.cov(self.mod.trunc_error)
        self.basis = self.mod.basis
        self.meas_error_cor = np.eye(self.basis.shape[0])
        self.discrep_cov = np.eye(self.basis.shape[0]) * 1e-12
        self.ii = 0
        npc = self.mod.nbasis
        self.mod_s2 = np.empty([self.nmcmc, npc])
        for i in range(npc):
            self.mod_s2[:, i] = self.mod.bm_list[i].samples.sdResid ** 2
        self.emu_vars = self.mod_s2[self.ii]
        self.yobs = None
        self.marg_lik_cov = None
        self.nd = 0
        if exp_ind is None:
            exp_ind = np.array(0)
        self.nexp = exp_ind.max() + 1
        self.exp_ind = exp_ind
        self.s2 = s2
        self.constants = None
        if s2 == "gibbs":
            raise "Cannot use Gibbs s2 for emulator models."
        return

    def step(self):
        self.ii = np.random.choice(range(self.nmcmc), 1).item()
        self.emu_vars = self.mod_s2[self.ii]
        return

    def eval(self, parmat, pool=None, nugget=False):
        """
        parmat : ~
        """
        parmat_array = np.vstack([
            parmat[v] for v in self.input_names
        ]).T  # get correct subset/ordering of inputs
        pred = self.mod.predict(
            parmat_array, mcmc_use=np.array([self.ii]), nugget=nugget
        )[0, :, :]

        if pool is True:
            return pred
        else:
            nrep = list(parmat.values())[0].shape[0] // self.nexp
            return np.concatenate(
                [
                    pred[
                        np.ix_(
                            np.arange(i, nrep * self.nexp, self.nexp),
                            np.where(self.exp_ind == i)[0],
                        )
                    ]
                    for i in range(self.nexp)
                ],
                1,
            )
            # this is evaluating all experiments for all thetas, which is overkill

    def llik(self, yobs, pred, cov):
        vec = yobs - pred
        out = -0.5 * (cov["ldet"] + vec.T @ cov["inv"] @ vec)
        return out

    def lik_cov_inv(self, s2vec):
        n = len(s2vec)
        Sigma = cor2cov(
            self.meas_error_cor[:n, :n], np.sqrt(s2vec)
        )  # :n is a hack for when ntheta>1 in heir...fix this sometime
        mat = (
            Sigma
            + self.trunc_error_cov
            + self.discrep_cov
            + self.basis @ np.diag(self.emu_vars) @ self.basis.T
        )
        # this doesnt work for vectorized experiments...maybe dont allow those for BASS
        chol = cholesky(mat)
        ldet = 2 * np.sum(np.log(np.diag(chol)))
        # la.dpotri(chol, overwrite_c=True) # overwrites chol with original matrix inverse
        inv = np.linalg.inv(mat)
        out = {"inv": inv, "ldet": ldet}
        return out


#######
### ModelBassPca_func: Model with BASS Emulator from pyBASS with Functional Response
class ModelBassPca_func(AbstractModel):
    """
    ModelBassPca_func: PCA Based BASS Model Emulator for Functional Outputs

    ModelBassPca_func Handles larger-dimensional functional responses (e.g., on large spatial fields) using
    various inversion tricks. We require any other covariance e.g., from discrepancy, measurement error, and basis truncation error)
    to be diagnonal. Smaller-dimensional functional responses could be specified with non-diagonal covariances using ModeBassPca_mult.
    """

    def __init__(self, bmod, input_names, exp_ind=None, s2="MH"):
        """
        bmod        : bassPCA fit
        input_names : list of the names of the inputs to bmod
        s2          : method for handling experiment-specific noise s2; options are 'MH' (Metropolis-Hastings Sampling), 'fix' (fixed at s2_est from addVecExperiments call)
        """
        self.mod = bmod
        self.stochastic = True
        self.nmcmc = len(bmod.bm_list[0].samples.s2)
        self.input_names = input_names
        self.basis = self.mod.basis
        self.meas_error_cor = np.eye(self.basis.shape[0])
        self.discrep_cov = np.eye(self.basis.shape[0]) * 1e-12
        self.ii = 0
        npc = self.mod.nbasis
        if npc > 1:
            self.trunc_error_var = np.diag(np.cov(self.mod.trunc_error))
        else:
            self.trunc_error_var = np.diag(
                np.cov(self.mod.trunc_error).reshape([1, 1])
            )
        self.mod_s2 = np.empty([self.nmcmc, npc])
        for i in range(npc):
            self.mod_s2[:, i] = self.mod.bm_list[i].samples.s2
        self.emu_vars = self.mod_s2[self.ii]
        self.yobs = None
        self.marg_lik_cov = None
        self.discrep_vars = None
        self.nd = 0
        self.discrep_tau = 1.0
        self.D = None
        self.discrep = 0.0
        if exp_ind is None:
            exp_ind = np.array(0)
        self.nexp = exp_ind.max() + 1
        self.exp_ind = exp_ind
        self.s2 = s2
        self.constants = None
        if s2 == "gibbs":
            raise "Cannot use Gibbs s2 for emulator models."
        return

    def step(self):
        self.ii = np.random.choice(range(self.nmcmc), 1).item()
        self.emu_vars = self.mod_s2[self.ii]
        return

    # @profile
    def discrep_sample(self, yobs, pred, cov, itemp):
        # if self.nd>0:
        S = np.linalg.inv(
            np.eye(self.nd) / self.discrep_tau + self.D.T @ cov["inv"] @ self.D
        )
        m = self.D.T @ cov["inv"] @ (yobs - pred)
        discrep_vars = chol_sample(S @ m, S / itemp)
        # self.discrep = self.D @ self.discrep_vars
        return discrep_vars

    # @profile
    def eval(self, parmat, pool=None, nugget=False):
        """
        parmat : ~
        """
        parmat_array = np.vstack([
            parmat[v] for v in self.input_names
        ]).T  # get correct subset/ordering of inputs
        pred = self.mod.predict(
            parmat_array, mcmc_use=np.array([self.ii]), nugget=nugget
        )[0, :, :]

        if pool is True:
            return pred
        else:
            nrep = list(parmat.values())[0].shape[0] // self.nexp
            return np.concatenate(
                [
                    pred[
                        np.ix_(
                            np.arange(i, nrep * self.nexp, self.nexp),
                            np.where(self.exp_ind == i)[0],
                        )
                    ]
                    for i in range(self.nexp)
                ],
                1,
            )
            # this is evaluating all experiments for all thetas, which is overkill

    # @profile
    def llik(self, yobs, pred, cov):
        vec = yobs - pred
        out = -0.5 * (cov["ldet"] + vec.T @ cov["inv"] @ vec)
        return out

    # @profile
    def lik_cov_inv(self, s2vec):
        vec = self.trunc_error_var + s2vec
        # mat = np.diag(vec) + self.basis @ np.diag(self.emu_vars) @ self.basis.T
        # inv = np.linalg.inv(mat)
        # ldet = np.linalg.slogdet(mat)[1]
        # out = {'inv' : inv, 'ldet' : ldet}
        Ainv = np.diag(1 / vec)
        Aldet = np.log(vec).sum()
        out = self.swm(
            Ainv,
            self.basis,
            np.diag(1 / self.emu_vars),
            self.basis.T,
            Aldet,
            np.log(self.emu_vars).sum(),
        )
        return out

    # @profile
    def chol_solve(self, x):
        mat = cho_factor(x)
        ldet = 2 * np.sum(np.log(np.diag(mat[0])))
        ##la.dpotri(mat, overwrite_c=True) # overwrites mat with original matrix inverse, but not correct
        # inv = cho_solve(mat, np.eye(x.shape[0])) # correct, but slower for small dimension
        inv = np.linalg.inv(x)
        out = {"inv": inv, "ldet": ldet}
        return out

    # @profile
    def swm(
        self, Ainv, U, Cinv, V, Aldet, Cldet
    ):  # sherman woodbury morrison (A+UCV)^-1 and |A+UCV|
        in_mat = self.chol_solve(Cinv + V @ Ainv @ U)
        inv = Ainv - Ainv @ U @ in_mat["inv"] @ V @ Ainv
        ldet = in_mat["ldet"] + Aldet + Cldet
        out = {"inv": inv, "ldet": ldet}
        return out


#######
### ModelBpprPca_func: Model with BayesPPR Emulator with Functional Responses
class ModelBpprPca_func(AbstractModel):
    """
    ModelBpprPca_func: BayesPPR Model Emulator for Functional Outputs

    ModelBpprPca_func Handles larger-dimensional functional responses (e.g., on large spatial fields) using
    various inversion tricks. We require any other covariance e.g., from discrepancy, measurement error, and basis truncation error)
    to be diagnonal. Smaller-dimensional functional responses could be specified with non-diagonal covariances using ModelBpprPca_mult.
    """

    def __init__(self, bmod, input_names, exp_ind=None, s2="MH"):
        """
        bmod        : bassPCA fit
        input_names : list of the names of the inputs to bmod
        s2          : method for handling experiment-specific noise s2; options are 'MH' (Metropolis-Hastings Sampling), 'fix' (fixed at s2_est from addVecExperiments call)
        """
        self.mod = bmod
        self.stochastic = True
        self.nmcmc = len(bmod.bm_list[0].samples.sdResid)
        self.input_names = input_names
        self.basis = self.mod.basis
        self.meas_error_cor = np.eye(self.basis.shape[0])
        self.discrep_cov = np.eye(self.basis.shape[0]) * 1e-12
        self.ii = 0
        npc = self.mod.nbasis
        if npc > 1:
            self.trunc_error_var = np.diag(np.cov(self.mod.trunc_error))
        else:
            self.trunc_error_var = np.diag(
                np.cov(self.mod.trunc_error).reshape([1, 1])
            )
        self.mod_s2 = np.empty([self.nmcmc, npc])
        for i in range(npc):
            self.mod_s2[:, i] = self.mod.bm_list[i].samples.sdResid ** 2
        self.emu_vars = self.mod_s2[self.ii]
        self.yobs = None
        self.marg_lik_cov = None
        self.discrep_vars = None
        self.nd = 0
        self.discrep_tau = 1.0
        self.D = None
        self.discrep = 0.0
        if exp_ind is None:
            exp_ind = np.array(0)
        self.nexp = exp_ind.max() + 1
        self.exp_ind = exp_ind
        self.s2 = s2
        self.constants = None
        if s2 == "gibbs":
            raise "Cannot use Gibbs s2 for emulator models."
        return

    def step(self):
        self.ii = np.random.choice(range(self.nmcmc), 1).item()
        self.emu_vars = self.mod_s2[self.ii]
        return

    # @profile
    def discrep_sample(self, yobs, pred, cov, itemp):
        # if self.nd>0:
        S = np.linalg.inv(
            np.eye(self.nd) / self.discrep_tau + self.D.T @ cov["inv"] @ self.D
        )
        m = self.D.T @ cov["inv"] @ (yobs - pred)
        discrep_vars = chol_sample(S @ m, S / itemp)
        # self.discrep = self.D @ self.discrep_vars
        return discrep_vars

    # @profile
    def eval(self, parmat, pool=None, nugget=False):
        """
        parmat : ~
        """
        parmat_array = np.vstack([
            parmat[v] for v in self.input_names
        ]).T  # get correct subset/ordering of inputs
        pred = self.mod.predict(
            parmat_array, mcmc_use=np.array([self.ii]), nugget=nugget
        )[0, :, :]

        if pool is True:
            return pred
        else:
            nrep = list(parmat.values())[0].shape[0] // self.nexp
            return np.concatenate(
                [
                    pred[
                        np.ix_(
                            np.arange(i, nrep * self.nexp, self.nexp),
                            np.where(self.exp_ind == i)[0],
                        )
                    ]
                    for i in range(self.nexp)
                ],
                1,
            )
            # this is evaluating all experiments for all thetas, which is overkill

    # @profile
    def llik(self, yobs, pred, cov):
        vec = yobs - pred
        out = -0.5 * (cov["ldet"] + vec.T @ cov["inv"] @ vec)
        return out

    # @profile
    def lik_cov_inv(self, s2vec):
        vec = self.trunc_error_var + s2vec
        # mat = np.diag(vec) + self.basis @ np.diag(self.emu_vars) @ self.basis.T
        # inv = np.linalg.inv(mat)
        # ldet = np.linalg.slogdet(mat)[1]
        # out = {'inv' : inv, 'ldet' : ldet}
        Ainv = np.diag(1 / vec)
        Aldet = np.log(vec).sum()
        out = self.swm(
            Ainv,
            self.basis,
            np.diag(1 / self.emu_vars),
            self.basis.T,
            Aldet,
            np.log(self.emu_vars).sum(),
        )
        return out

    # @profile
    def chol_solve(self, x):
        mat = cho_factor(x)
        ldet = 2 * np.sum(np.log(np.diag(mat[0])))
        ##la.dpotri(mat, overwrite_c=True) # overwrites mat with original matrix inverse, but not correct
        # inv = cho_solve(mat, np.eye(x.shape[0])) # correct, but slower for small dimension
        inv = np.linalg.inv(x)
        out = {"inv": inv, "ldet": ldet}
        return out

    # @profile
    def swm(
        self, Ainv, U, Cinv, V, Aldet, Cldet
    ):  # sherman woodbury morrison (A+UCV)^-1 and |A+UCV|
        in_mat = self.chol_solve(Cinv + V @ Ainv @ U)
        inv = Ainv - Ainv @ U @ in_mat["inv"] @ V @ Ainv
        ldet = in_mat["ldet"] + Aldet + Cldet
        out = {"inv": inv, "ldet": ldet}
        return out


#######
### ModelF: Function for Simulator Model Evaluation or Evaluation of Alternative Emulator Model
class ModelF(AbstractModel):
    """Custom Simulator/Emulator Model"""

    def __init__(
        self, f, input_names, exp_ind=None, s2="gibbs"
    ):  # not sure if this is vectorized
        """
        f           : user-defined function taking single input with elements x[0] = first element of theta, x[1] = second element of theta, etc. Function must output predictions for all observations
        input_names : list of the names of the inputs to bmod
        s2          : method for handling experiment-specific noise s2; options are 'MH' (Metropolis-Hastings Sampling), 'fix' (fixed at s2_est from addVecExperiments call), and 'gibbs' (Gibbs sampling)
        """
        self.mod = f
        self.input_names = input_names
        self.stochastic = False
        self.yobs = None
        self.meas_error_cor = 1.0  # np.diag(self.basis.shape[0])
        if exp_ind is None:
            exp_ind = np.array(0)
        self.nexp = exp_ind.max() + 1
        self.exp_ind = exp_ind
        self.nd = 0
        self.s2 = s2
        self.constants = None

    def eval(self, parmat, pool=None, nugget=False):
        parmat_array = np.vstack([
            parmat[v] for v in self.input_names
        ]).T  # get correct subset/ordering of inputs
        if pool is True:
            return np.apply_along_axis(self.mod, 1, parmat_array)
        else:
            nrep = list(parmat.values())[0].shape[0] // self.nexp
            out_all = np.apply_along_axis(self.mod, 1, parmat_array)

            # out_sub = np.concatenate([out_all[(i*nrep):(i*nrep+nrep), self.exp_ind==i] for i in range(self.nexp)], 1)
            out_sub = np.concatenate(
                [
                    out_all[
                        np.ix_(
                            np.arange(i, nrep * self.nexp, self.nexp),
                            np.where(self.exp_ind == i)[0],
                        )
                    ]
                    for i in range(self.nexp)
                ],
                1,
            )
            return out_sub
            # this is evaluating all experiments for all thetas, which is overkill
        # need to have some way of dealing with non-pooled eval fo this and bassPCA version

    def discrep_sample(
        self, yobs, pred, cov, itemp
    ):  # Added by Lauren on 11/17/23.
        S = np.linalg.inv(
            np.eye(self.nd) / self.discrep_tau  # defined by addVecExperiments
            + self.D.T @ (cov["inv"].flatten() * np.eye(len(yobs))) @ self.D
        )
        m = self.D.T @ (cov["inv"] * np.eye(len(yobs))) @ (yobs - pred)
        discrep_vars = chol_sample(S @ m, S / itemp)
        return discrep_vars


#######
### ModelF_bigdata: Function for Simulator Model Evaluation or Evaluation of Alternative Emulator Model using Bigger Data
class ModelF_bigdata(AbstractModel):
    """Custom Simulator/Emulator Model"""

    def __init__(
        self, f, input_names, exp_ind=None, s2="gibbs"
    ):  # not sure if this is vectorized
        """
        f           : user-defined function taking single input with elements x[0] = first element of theta, x[1] = second element of theta, etc. Function must output predictions for all observations
        input_names : list of the names of the inputs to bmod
        s2          : method for handling experiment-specific noise s2; options are 'MH' (Metropolis-Hastings Sampling), 'fix' (fixed at s2_est from addVecExperiments call), and 'gibbs' (Gibbs sampling)
        """
        self.mod = f
        self.input_names = input_names
        self.stochastic = False
        self.yobs = None
        self.meas_error_cor = 1.0  # np.diag(self.basis.shape[0])
        if exp_ind is None:
            exp_ind = np.array(0)
        self.nexp = exp_ind.max() + 1
        self.exp_ind = exp_ind
        self.nd = 0
        self.s2 = s2
        self.vec = np.linspace(
            1, 16200, 16200
        )  # define to speed up llik evaluation
        self.vec2 = np.linspace(
            1, 16200, 16200
        )  # define to speed up llik evaluation
        self.inv = np.linspace(
            1, 16200, 16200
        )  # define to speed up lik_cov_inv evaluation
        self.m = np.linspace(
            1, 16200, 16200
        )  # define to speed up discrep_sample evaluation
        self.vmat = np.linspace(
            1, 16200, 16200
        )  # define to speed up discrep_sample evaluation
        self.constants = None

    def eval(self, parmat, pool=None, nugget=False):
        parmat_array = np.vstack([
            parmat[v] for v in self.input_names
        ]).T  # get correct subset/ordering of inputs
        if pool is True:
            return np.apply_along_axis(self.mod, 1, parmat_array)
        else:
            nrep = list(parmat.values())[0].shape[0] // self.nexp
            out_all = np.apply_along_axis(self.mod, 1, parmat_array)

            # out_sub = np.concatenate([out_all[(i*nrep):(i*nrep+nrep), self.exp_ind==i] for i in range(self.nexp)], 1)
            out_sub = np.concatenate(
                [
                    out_all[
                        np.ix_(
                            np.arange(i, nrep * self.nexp, self.nexp),
                            np.where(self.exp_ind == i)[0],
                        )
                    ]
                    for i in range(self.nexp)
                ],
                1,
            )
            return out_sub
            # this is evaluating all experiments for all thetas, which is overkill
        # need to have some way of dealing with non-pooled eval fo this and bassPCA version

    def discrep_sample(
        self, yobs, pred, cov, itemp
    ):  # Added by Lauren on 11/17/23.
        self.vec = cov["inv"].reshape(-1, 1) * (yobs - pred).reshape(-1, 1)
        self.vmat = (
            np.repeat(cov["inv"].reshape(-1, 1), self.nd, axis=1) * self.D
        )
        S = np.linalg.inv(
            # np.eye(self.nd) / self.discrep_tau #defined by addVecExperiments
            np.diag(
                1 / self.discrep_tau
            )  # modification to allow vector-valued discrep_tau
            + self.D.T @ self.vmat
        )
        self.m = self.D.T @ self.vec
        discrep_vars = chol_sample((S @ self.m).flatten(), S / itemp)
        return discrep_vars

    def llik(self, yobs, pred, cov):  # assumes diagonal cov
        self.vec = yobs.flatten() - pred.flatten()
        self.vec2 = self.vec * self.vec * cov["inv"]
        out = -0.5 * cov["ldet"] - 0.5 * self.vec2.sum()
        return out

    def lik_cov_inv(self, s2vec):  # default is diagonal covariance matrix
        self.inv = 1 / s2vec
        ldet = np.log(s2vec).sum()
        out = {"inv": self.inv, "ldet": ldet}
        return out


#######
### ModelMaterialStrength: PTW Model for Hopi-Bar / Quasistatic Experiments
epsilon = 1e-5  # used in ModelMaterialStrength to pad set of strains for evaluate for each experiment, user can ignore


class ModelMaterialStrength(AbstractModel):
    """
    PTW Model for Hoppy-Bar / Quasistatic Experiments

    Currently not able to handle pooled model discrepancy.
    """

    def __init__(
        self,
        temps,
        edots,
        consts,
        strain_histories,
        flow_stress_model,
        melt_model,
        shear_model,
        specific_heat_model,
        density_model,
        pool=True,
        s2="gibbs",
    ):
        """
        temps               : list of temperatures indexed by experiment (units = Kelvin)
        edots               : edots: list of strain rates indexed by experiment (units = NEED TO ADD)
        consts              : dictionary of constants for PTW model. Use showdef_ModelMaterialStrength(str_func_name) to figure out the needed constants
        strain_histories    : List of strain histories for HB/Quasistatic Experiments
        flow_stress_model   : options provided by getoptions_ModelMaterialStrength()['flow_stress_model']
        melt_model          : options provided by getoptions_ModelMaterialStrength()['melt_model']
        shear_model         : options provided by getoptions_ModelMaterialStrength()['shear_model']
        specific_heat_model : options provided by getoptions_ModelMaterialStrength()['specific_heat_model']
        density_model       : options provided by getoptions_ModelMaterialStrength()['density_model']
        pool                : False if fitting hierarchical model, True if fitting pooled model
        s2                  : method for handling experiment-specific noise s2; options are 'MH' (Metropolis-Hastings Sampling), 'fix' (fixed at s2_est from addVecExperiments call)
        """
        self.meas_strain_histories = strain_histories
        self.meas_strain_max = np.array([v.max() for v in strain_histories])
        self.strain_max = self.meas_strain_max.max()
        self.nhists = sum([len(v) for v in strain_histories])
        self.model = pm_vec.MaterialModel(
            flow_stress_model=eval("pm_vec." + flow_stress_model),
            shear_modulus_model=eval("pm_vec." + shear_model),
            specific_heat_model=eval("pm_vec." + specific_heat_model),
            melt_model=eval("pm_vec." + melt_model),
            density_model=eval("pm_vec." + density_model),
        )
        self.model_info = [
            flow_stress_model,
            shear_model,
            specific_heat_model,
            melt_model,
            density_model,
        ]
        self.constants = consts
        self.temps = temps
        self.edots = edots
        self.nexp = len(strain_histories)
        self.Nhist = 100
        self.stochastic = False
        self.pool = pool
        self.yobs = None
        self.nd = 0
        self.s2 = s2

        # self.meas_error_cor = np.diag(self.basis.shape[0])
        return

    def eval(
        self, parmat, pool=None, nugget=False
    ):  # note: extra parameters ignored
        """parmat:  dictionary of parameters"""
        if (pool is True) or self.pool:  # Pooled Case
            # nrep = parmat['p'].shape[0]  # number of temper temps
            nrep = list(parmat.values())[0].shape[0]
            parmat_big = {
                key: np.kron(parm, np.ones(self.nexp))
                for key, parm in parmat.items()
            }
        else:  # hierarchical case
            nrep = (
                list(parmat.values())[0].shape[0] // self.nexp
            )  # number of temper temps
            parmat_big = parmat

        edots = np.kron(
            np.ones(nrep), self.edots
        )  # 1d vector, nexp * temper_temps
        temps = np.kron(
            np.ones(nrep), self.temps
        )  # 1d vector, nexp * temper_temps
        strain_maxs = np.kron(
            np.ones(nrep), self.meas_strain_max
        )  # 1d vector, nexp * temper_temps
        ntot = edots.shape[0]  # nexp * temper_temps
        sim_strain_histories = pm_vec.generate_strain_history_new(
            strain_maxs, edots, self.Nhist
        )
        self.model.initialize(parmat_big, self.constants)
        self.model.initialize_state(
            T=temps, stress=np.zeros(ntot), strain=np.zeros(ntot)
        )
        sim_state_histories = self.model.compute_state_history(
            sim_strain_histories
        )
        sim_strains = sim_state_histories[:, 1].T  # 2d array: ntot, Nhist
        sim_stresses = sim_state_histories[:, 2].T  # 2d array: ntot, Nhist

        # Expand/Flatten Simulated strains to single vector--ensure no overlap.
        strain_ends = np.hstack((
            0.0,
            np.cumsum(strain_maxs + epsilon)[:-1],
        ))  # 1d vector, ntot
        flattened_sim_strain = np.hstack(  # 1d vector: ntot * Nhist
            [x + y for x, y in zip(sim_strains, strain_ends)]
        )
        # Expand/flatten simulated stress to single vector
        flattened_sim_stress = np.hstack(sim_stresses)
        # Expand/flatten measured strain to single vector, for each parameter.  Use same
        #  Computed strain ends to ensure no overlap
        flattened_strain = (
            np.hstack(  # cycle will repeat through measured strain histories
                [
                    x + y
                    for x, y in zip(
                        cycle(self.meas_strain_histories), strain_ends
                    )
                ]
            )
        )
        ifunc = interp1d(  # Generate the interpolation function.
            flattened_sim_strain,
            flattened_sim_stress,
            kind="linear",
            assume_sorted=True,
        )
        ypred = ifunc(flattened_strain).reshape(
            nrep, -1
        )  # Interpolate, and output.
        return ypred


def interpolate_experiment(args):
    """Interpolate and predict at x.  Args is tuple(x_observed, y_observed, x_new)"""
    ifunc = interp1d(args[0], args[1], kind="cubic")
    return ifunc(args[2])


#######
### getoptions_ModelMaterialStrength: Provides current options for ModelMaterialStrength physical models
def getoptions_ModelMaterialStrength():
    import re

    import impala

    mod_options = dir(impala.physics.physical_models_vec)
    flow_stress_model = list(
        filter(re.compile(".*Yield_Stress").match, mod_options)
    )
    flow_stress_model = np.append(
        flow_stress_model,
        list(filter(re.compile(".*Flow_Stress").match, mod_options)),
    )
    flow_stress_model = list(flow_stress_model)
    melt_model = list(
        filter(re.compile(".*Melt_Temperature").match, mod_options)
    )
    shear_model = list(filter(re.compile(".*Shear_Modulus").match, mod_options))
    specific_heat_model = list(
        filter(re.compile(".*Specific_Heat").match, mod_options)
    )
    density_model = list(filter(re.compile(".*Density").match, mod_options))
    return dict({
        "flow_stress_model": flow_stress_model,
        "melt_model": melt_model,
        "shear_model": shear_model,
        "specific_heat_model": specific_heat_model,
        "density_model": density_model,
    })


#######
### showdef_ModelMaterialStrength: Shows the definition for a given ModelMaterialStrength function listed in getoptions_ModelMaterialStrength()
def showdef_ModelMaterialStrength(func_name):
    """
    func_name: string listing a function listed in get_ModelMaterialStrength_options(), e.g., showdef_ModelMaterialStrength('Linear_Specific_Heat')
    """
    import inspect

    import impala

    my_func = getattr(impala.physics, func_name)
    print(inspect.getsource(my_func))


# EOF



"""
Basic progress bar.
"""

from datetime import timedelta, datetime
import time
import sys

def pbrange(*args, **kwargs):
    """
    # Example 1:
    for i in pbrange(n):
        time.sleep(s)

    # Example 2: Context and add extra args, using prange.
    print("pbrange in context:")
    with pbrange(0, 5 * n, 5) as pb:
        for i in pb:
            pb.extra = {f"{i}^2": i ** 2}
            time.sleep(s)
    """

    return pbar(range(*args), **kwargs)

class pbar:
    """
    # Example 1:
    for i in pbar(range(n)):
        time.sleep(s)

    # Example 2:
    pb = pbar(range(n))
    for i in pb:
        time.sleep(s)

    # Example 3: Context and add extra args.
    xs = range(0, 5 * n, 5)
    with pbar(xs) as pb:
        for i in pb:
            pb.extra = {f"{i}^2": i ** 2}
            time.sleep(s)
    """
    def __init__(self, iterable, min_interval=0.2, show=lambda: True, fname=None):
        self.iterator = iterable.__iter__()
        self.iters = iterable.__len__()
        self.min_interval = min_interval
        self.genesis = time.time()
        self.tic = self.genesis
        self.extra = None
        self.show = show
        self.fname = fname
        self.stdout = sys.stdout
        self.set_stdout()

    def parse_extra(self):
        if self.extra is None:
            return ""
        else:
            return " | " + " | ".join(f"{name}: {value}" for name, value in self.extra.items())

    def clean(self, x):
        return str(x).split(".")[0]

    def __iter__(self):
        self.i = 0
        return self

    def __next__(self):
        toc = time.time()
        if self.show():
            self.print(toc)
        self.i += 1
        if self.i > self.iters and self.show():
            if self.fname is None:
                print()
        return self.iterator.__next__()

    def set_stdout(self):
        if self.fname is not None:
            sys.stdout = open(self.fname, "w")

    def print(self, toc):
        if self.i == 0:
            self.bar = (
                f"{self.clean(datetime.now()) }"
            )
            print(f"[{self.bar}]", end="\r", flush=True)
        elif toc - self.tic > self.min_interval or self.i == self.iters:
            elapsed = toc - self.genesis
            speed = f"{self.i / elapsed:.2f}" if self.i > 0 else "???"
            wall = timedelta(seconds=elapsed)
            eta = wall * (self.iters / self.i - 1)
            perc = int(100 * self.i / self.iters)
            self.bar = (
                f"{self.clean(datetime.now()) } | "
                f"{self.i}/{self.iters} ({perc}%) | "
                f"WALL: {self.clean(wall)} | "
                f"ETA: {self.clean(eta)} | "
                f"{speed}it/s" + self.parse_extra()
            )
            print(f"[{self.bar}]", end="\r", flush=True)
            self.tic = toc

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        sys.stdout = self.stdout


